#!/usr/bin/env python
"""Utility functions for rocks."""

import collections
from functools import reduce
import json
import keyword
import os
import pickle
import sys
import time
import warnings

import click
from iterfzf import iterfzf
import numpy as np
import pandas as pd
import requests
from tqdm import tqdm

import rocks


# ------
# Index functions
def read_index():
    """Read local index of asteroid numbers, names, SsODNet IDs.

    Returns
    =======
    pd.DataFrame
        Asteroid index with three columns.
    """
    with open(rocks.PATH_INDEX, "rb") as ind:
        return pickle.load(ind)


def create_index():
    """Update index of numbered SSOs."""

    # Get currently indexed objects
    index = read_index()

    # Get list of numbered asteroids from MPC
    url = "https://www.minorplanetcenter.net/iau/lists/NumberedMPs.txt"
    numbered = pd.read_fwf(url, colspecs=[(0, 7)], names=["number"])
    numbered = set(int(n.strip(" (")) for n in numbered.number)  # type: ignore

    # Compare list to index
    missing = set(index.number) ^ set(numbered)

    if not missing:
        return

    # Get ids of missing entries, append to index
    names, numbers, ids = zip(*rocks.identify(missing))

    index = index.append(
        pd.DataFrame(
            {
                "name": names,
                "number": numbers,
                "id_": ids,
            }
        )
    )

    # Save index to file
    index = (
        index.dropna(how="any")
        .drop_duplicates("number")
        .sort_values("number", ascending=True)
        .astype({"number": np.int64, "name": str, "id_": str})
    )

    with open(rocks.PATH_INDEX, "wb") as ind:
        pickle.dump(index, ind, protocol=4)


def select_sso_from_index():  # pragma: no cover
    """Select SSO numbers and designations from interactive fuzzy search.

    Returns
    =======
    str
        asteroid name or designation
    int
        asteroid number

    Notes
    -----
    If the selection is interrupted with ctrl-c, ``(None, None)`` is returned.

    Examples
    --------
    >>> import rocks
    >>> name, number, id_ = rocks.utils.select_sso_from_index()
    """
    INDEX = read_index()

    def _fuzzy_desig_selection(index):
        """Generator for fuzzy search of asteroid index file. """
        for number, name in zip(index.number, index["name"]):
            yield f"{number} {name}"

    try:
        nunaid = iterfzf(_fuzzy_desig_selection(INDEX), exact=True)
        number, *name = nunaid.split()

    except AttributeError:  # no SSO selected
        return None, np.nan, None

    id_ = INDEX.loc[INDEX.number == int(number), "id_"].values[0]

    return " ".join(name), int(number), id_


# ------
# ssoCard utility functions
def rsetattr(obj, attr, val):  # pragma: no cover
    """Deep version of setattr."""
    pre, _, post = attr.rpartition(".")
    return setattr(rgetattr(obj, pre) if pre else obj, post, val)


def rgetattr(obj, attr, *args):  # pragma: no cover
    """Deep version of getattr."""

    def _getattr(obj, attr):
        return getattr(obj, attr, *args)

    return reduce(_getattr, [obj] + attr.split("."))


def update_ssoCard(dict_, update):
    """Recursively update ssoCard template with retrieved values.

    Parameters
    ==========
    dict_ : dict, list
        The values in the template or retrieved ssoCard.
    update : dict, list
        The values in the retrieved ssoCard.

    Returns
    =======
    dict
        The updated ssoCard.
    """
    for key, val in update.items():
        if isinstance(val, collections.abc.Mapping):  # type: ignore
            dict_[key] = update_ssoCard(dict_.get(key, {}), val)
        else:
            if isinstance(dict_, list):  # multiple pair members
                dict_ = dict_[-1]  # pragma: no cover
            if isinstance(val, list):  # multiple taxonomies
                val = val[-1]
            dict_[key] = val
    return dict_


def create_ssoCard_template():  # pragma: no cover
    """Retrieve current ssoCard template from SsODNet and save
    simplified version to file.
    """
    response = requests.get(
        "https://ssp.imcce.fr/webservices/ssodnet/api/"
        "ssocard/templates/catalogue-template_aster-astorb.json"
    )

    ssoCard = response.json()
    ssoCard = set_endpoints_to_nan(ssoCard)
    ssoCard = sanitize_keys(ssoCard)

    with open(rocks.PATH_TEMPLATE, "w") as file_:
        json.dump(ssoCard, file_)


def sanitize_keys(dict_):
    """Recursively ensure that dict keys are not python keywords."""
    for key, value in dict_.copy().items():
        if isinstance(value, dict):
            dict_[key] = sanitize_keys(dict_[key])
        elif isinstance(value, list):
            if all([isinstance(v, dict) for v in value]):
                dict_[key] = [sanitize_keys(v) for v in value]
        if keyword.iskeyword(key):
            dict_[f"{key}_"] = dict_.pop(key)
    return dict_


def set_endpoints_to_nan(dict_):
    """Recursively sets dict values to NaN if they are not of type dict or unit."""
    for key, value in dict_.copy().items():
        if isinstance(value, dict):
            dict_[key] = set_endpoints_to_nan(dict_[key])
        else:
            if key == "unit":
                continue
            else:
                dict_[key] = np.nan
    return dict_


# ------
# SsODNet functions
def get_ssoCard(id_, only_cache=False):
    """Retrieve single ssoCard either from cache or SsODNet.

    Parameters
    ==========
    id_ : str
        Single Minor body target id from SsODNet.
    only_cache : bool
        Do not query SsODNet, return None instead.

    Returns
    =======
    dict, None
        Single ssoCard in json format if successful.
        None if query failed or not queried.
    """
    if not isinstance(id_, str):
        warnings.warn(f"Expected string identifier, got {type(id_)}.")
        return None

    # Check presence in cache
    PATH_CARD = os.path.join(rocks.PATH_CACHE, f"{id_}.json")

    if not os.path.isfile(PATH_CARD):
        if not only_cache:
            return __query_ssoCards(id_)[0]
        else:
            return None
    else:
        with open(PATH_CARD, "r") as file_:
            ssoCard = json.load(file_)
        return ssoCard


def get_ssoCards(ids, progress=False):
    """Return target ssoCard. Use cache if existant, else, query SsODNet
    and cache results. Accepts multiple ids.

    Parameters
    ==========
    ids : str, list, np.array, pd.Series
        Single Minor body target id from SsODNet or list of several.
        Pass SsODNet ID for fast access.
    progress : bool
        Show progress of instantiation. Default is False.

    Returns
    =======
    dict, list of dict, None
        Single or list of ssoCards in json format if successful.
        None if query failed.
    """
    if isinstance(ids, str):
        ids = [ids]
    elif isinstance(ids, pd.Series):
        ids = ids.values

    id_card = collections.OrderedDict(
        (id_, get_ssoCard(id_, only_cache=True)) for id_ in ids
    )
    missing = [id_ for id_, card in id_card.items() if card is None]

    if not missing:
        return id_card.values()

    print(f"Retrieving {len(missing)} ssoCards..")

    # Perform POST query for missing cards in chunks
    if progress:
        progressbar = tqdm(desc="Retrieving ssoCards", total=len(missing))
    else:
        progressbar = False

    for subset in [missing[i : i + 500] for i in range(0, len(missing), 500)]:

        ssoCards = __query_ssoCards(subset, progressbar)

        for id_, ssoCard in zip(subset, ssoCards):

            if ssoCard is None:
                continue

            # if ssoCard is None, this value is already None
            id_card[id_] = ssoCard

            with open(f"{rocks.PATH_CACHE}/{id_}.json", "w") as file_:
                json.dump(ssoCard, file_)

    # Return requested ssoCards
    return id_card.values()


def __query_ssoCards(ids, progressbar=False):
    """Retrieve ssoCards for targets from SsODNet via POST request.

    Parameters
    ==========
    ids : str, list, np.array, pd.Series
        Single Minor body target id from SsODNet or list of several.
    progressbar : tqdm.tqdm
        A progress bar instance. Default is False, no progressbar.

    Returns
    =======
    list of dict, None
        List of ssoCards in json format if successful.
        None if query failed.

    Notes
    =====
    Performs output check and tries to infer correct target string if query failed.
    """
    if isinstance(ids, str):
        ids = [ids]
    elif isinstance(ids, pd.Series):
        ids = ids.values
    if any(pd.isnull(ids)):

        warnings.warn(
            f"There {'are' if sum(pd.isnull(ids)) > 1 else 'is a'} NaN"
            f" {'values' if sum(pd.isnull(ids)) > 1 else 'value'} among the IDs,"
            f" returning None in {'their' if sum(pd.isnull(ids)) > 1 else 'its'} place."
        )

        # Get indices of NaN values
        ind_nan = np.where(pd.isnull(ids))[0]

        for ind in ind_nan:
            ids.pop(ind)
    else:
        ind_nan = []

    url = "https://asterws2.imcce.fr/webservices/ssodnet/api/ssocard/"

    params = {"from": "rocks"}
    file_ = {"targets": ("\n".join([id_ for id_ in ids]))}

    try:
        response = requests.post(url=f"{url}mytargets", params=params, files=file_)
        response.encoding = "UTF-8"
        response = response.json()
    except requests.exceptions.RequestException as e:
        warnings.warn(f"An error occurred when retrieving the ssoCards: {e}")
        return None

    ssoCards = [response[id_] for id_ in ids]

    # Add False for the NaN indices removed above
    for ind in ind_nan:
        ssoCards.insert(ind, False)

    if progressbar:
        progressbar.update(n=len(ids))
    return ssoCards


def retrieve_catalogue(url):
    """Retrieve catalogue from datacloud."""
    response = requests.get(url.replace("SsODNetSsoCard", "rocks"))

    catalogue = response.json()["data"]
    catalogue = sanitize_keys(catalogue)

    return catalogue


# ------
# Pretty-print methods
def pretty_print_card(ssoCard, minimal):
    """Pretty-print ssoCard to terminal. Called by rocks info.

    Parameters
    ==========
    ssoCard : dict
        Minor body ssoCard.
    minimal : bool
        Reduce information to basics.
    """
    from rich.console import Console

    console = Console()

    if not minimal:
        # Print entire JSON
        console.print(ssoCard)
    else:

        from rich.panel import Panel
        from rocks.core import Rock

        rock = Rock(ssoCard["id"], ssoCard, skip_id_check=True)

        minimal_card = f"""

        type: {rock.type:<12} class: {rock.class_}
        taxonomy: {rock.taxonomy.class_:<12} ap: {rock.proper_elements.proper_semi_major_axis:.4}

        """  # {rock.proper_elements.semi_major_axis.unit}
        console.print(Panel(minimal_card, title=f"({rock.number}) {rock.name}"))


def pretty_print_property(rock, prop, plot=False):
    """Echo asteroid property for a single minor body.
    Print datacloud collections. Optionally open plots.

    Parameters
    ==========
    rock : rocks.Rock
        The asteroid Rock instance
    prop : str
        Asteroid property, attribute from JSON template
    plot : bool
        Plot propertyCollection. Default is False.
    """
    if isinstance(prop, rocks.core.stringParameter):
        _echo_stringParameter(prop)
    elif isinstance(prop, rocks.core.intParameter):
        _echo_intParameter(prop)
    elif isinstance(prop, rocks.core.floatParameter):
        _echo_floatParameter(prop)
    elif isinstance(prop, rocks.core.propertyCollection):
        _echo_propertyCollection(rock, prop, plot)
    elif isinstance(prop, rocks.core.listSameTypeParameter):
        _echo_listSameTyeParameter(prop, plot)
    else:
        print(prop)

    sys.exit()


def _echo_stringParameter(prop):
    """Echos stringParameter value."""
    print(prop)


def _echo_intParameter(prop):
    """Echos intParameter value."""
    if hasattr(prop, "unit"):
        unit = getattr(prop, "unit")
    else:
        unit = ""
    if hasattr(prop, "uncertainty"):
        uncert = getattr(prop, "uncertainty")
    else:
        uncert = ""

    if uncert:
        print(f"{prop:,} +- {uncert:,} {unit}")
    else:
        print(f"{prop}")


def _echo_floatParameter(prop):
    """Echos floatParameter value."""
    if hasattr(prop, "unit"):
        unit = getattr(prop, "unit")
    else:
        unit = ""
    if hasattr(prop, "uncertainty"):
        uncert = getattr(prop, "uncertainty")
    else:
        uncert = ""

    if uncert:
        print(f"{float(prop):.4} +- {float(uncert):.3} {unit}")
    else:
        print(f"{prop}")


def _echo_propertyCollection(rock, prop, plot):
    """Echos propertyCollection value.

    Parameters
    ==========
    rock : rocks.core.Rock
        The Rock instance.
    prop
        The attribute to print.
    """
    # TODO Add weighted average for floatParameters
    from rich import box
    from rich.table import Table
    from rich import print as rprint

    table = Table(
        header_style="bold blue",
        caption=f"({rock.number}) {rock.name}",
        box=box.SQUARE,
        footer_style="dim",
    )

    # Keys are table property names
    keys = [k for k in list(prop.__dict__.keys()) if k not in DONT_PRINT]

    for key in keys:
        table.add_column(key)

    if "diameters" in sys.argv:
        prop.preferred = prop.preferred_diameter
        prop_name = "diameter"
    elif "albedos" in sys.argv:
        prop.preferred = prop.preferred_albedo
        prop_name = "albedo"
    elif "masses" in sys.argv:
        prop_name = "mass"

    # Values are entries for each field
    for row in prop:
        table.add_row(
            *[str(getattr(row, key)) for key in keys],
            style="bold green"
            if hasattr(row, "preferred") and row.preferred
            else "white",
        )

    rprint(table)

    if plot:
        if any([p in sys.argv for p in ["-h", "--hist"]]):
            prop.hist(prop_name, show=True)
        if any([p in sys.argv for p in ["-s", "--scatter"]]):
            prop.scatter(prop_name, show=True)

    sys.exit()  # otherwise click prints Error


def _echo_listSameTyeParameter(prop):
    print(prop)


# ------
# Numerical methods
def weighted_average(observable, error):
    """Computes weighted average of observable.

    Parameters
    ----------
    observable : np.ndarray
        Float values of observable
    error : np.ndarray
        Corresponding errors of observable.

    Returns
    -------
    (float, float)
        Weighted average and its standard error.
    """
    if len(observable) == 1:
        return (observable[0], error[0])

    # Compute normalized weights
    weights = 1 / np.array(error) ** 2

    # Compute weighted average and uncertainty
    avg = np.average(observable, weights=weights)

    # Kirchner Case II
    # http://seismo.berkeley.edu/~kirchner/Toolkits/Toolkit_12.pdf
    var_avg = (
        len(observable)
        / (len(observable) - 1)
        * (
            sum(w * o ** 2 for w, o in zip(weights, observable)) / sum(weights)
            - avg ** 2
        )
    )
    std_avg = np.sqrt(var_avg / len(observable))
    return (avg, std_avg)


# ------
# Definitions
DATACLOUD_META = {
    # datacloud key : rocks name
    #     attr_name : Rock.xyz
    #     prop_name : Name of property in catalogue
    "aams": {
        "attr_name": "aams",
    },
    "astdys": {
        "attr_name": "astdys",
    },
    "astorb": {
        "attr_name": "astorb",
    },
    "binarymp_tab": {
        "attr_name": "binaries",
    },
    "diamalbedo": {
        "attr_name": "diamalbedo",
        "prop_name": {"diameters": "diameter", "albedos": "albedo"},
    },
    "families": {
        "attr_name": "families",
    },
    "masses": {
        "attr_name": "masses",
        "prop_name": {"masses": "mass"},
    },
    "mpcatobs": {
        "attr_name": "mpc",
    },
    "pairs": {
        "attr_name": "pairs",
    },
    "taxonomy": {
        "attr_name": "taxonomies",
    },
}

DONT_PRINT = [
    "num",
    "name",
    "id",
    "iddataset",
    "title",
    "url",
    "doi",
    "bibcode",
    "link",
    "datasetname",
    "idcollection",
    "selection",
    "source",
    "resourcename",
    "preferred",
    "preferred_albedo",
    "preferred_diameter",
    "_iter_index",
]



#!/usr/bin/env python
"""rocks command line suite."""
import os
import sys
import webbrowser

import click
import pandas as pd

import rocks


class AliasedGroup(click.Group):
    """Click group with custom default mode implementation. """

    def get_command(self, ctx, cmd_name):

        rv = click.Group.get_command(self, ctx, cmd_name)

        # ------
        # If it's a known subcommand, execute it
        if rv is not None:
            return rv

        # ------
        # Command aliases
        if cmd_name == "id":
            return identify

        # ------
        # Unknown subcommand, check if it's a valid property
        valid = False

        # ssoCard properties
        valid_props = [
            p for p in rocks.TEMPLATE_KEYS.columns if not p.startswith("datacloud")
        ]
        # missing intermediate levels
        valid_props = set(
            valid_props + [".".join(v.split(".")[:-1]) for v in valid_props if "." in v]
        )

        valid_props = sorted(list(valid_props), key=len)

        # Check if valid ssoCard property
        for prop in valid_props:

            if prop.endswith("class") or prop.endswith("complex"):
                prop += "_"

            if prop.endswith(cmd_name):
                valid = True
                datacloud = []
                break

        # datacloud properties
        if not valid:
            valid_catalogues = [
                (datacloud, cat["attr_name"])
                for datacloud, cat in rocks.utils.DATACLOUD_META.items()
            ]

            requested_catalogue = cmd_name.split(".")[0]

            if requested_catalogue in ["diameters", "albedos"]:
                requested_catalogue = "diamalbedo"

            for datacloud, cat in valid_catalogues:

                if cat == requested_catalogue:
                    valid = True
                    if len(cmd_name.split(".")) > 1:
                        # single property
                        prop = ".".join([cat, *cmd_name.split(".")[1:]])
                    else:
                        # catalogue overview
                        prop = cat
                    datacloud = [datacloud]
                    break

        if valid:

            arguments = sys.argv.copy()

            plot = False
            for p in ["-h", "--hist", "-s", "--scatter"]:
                if p in arguments:
                    plot = True
                    arguments.remove(p)

            # Identify rock and check datacloud catalogue
            if len(arguments) == 3:
                id_ = arguments[-1]
                skip_id_check = False
            else:
                _, _, id_ = rocks.utils.select_sso_from_index()
                skip_id_check = True

            rock = rocks.Rock(id_, datacloud=datacloud, skip_id_check=skip_id_check)
            if not isinstance(rock.id, str):
                sys.exit()

            prop = rocks.utils.rgetattr(rock, prop)
            return rocks.utils.pretty_print_property(rock, prop, plot=plot)
        else:
            return None


@click.group(cls=AliasedGroup)
def cli_rocks():
    """CLI suite for minor body exploration."""
    pass


@cli_rocks.command()
def docs():
    """Open rocks documentation in browser."""
    webbrowser.open("https://rocks.readthedocs.io/en/latest/", new=2)


@cli_rocks.command()
def update():
    """Update index of numbered SSOs and SsODNet metadata.

    The list is retrieved from the Minor Planet Center, at
    https://www.minorplanetcenter.net/iau/lists/NumberedMPs.txt
    """
    rocks.utils.create_index()
    rocks.utils.create_ssoCard_template()


@cli_rocks.command()
@click.argument("this", required=1)
def identify(this):
    """Get asteroid name and number from string input.

    Parameters
    ----------
    this : str
        String to identify asteroid.
    """
    name, number, _ = rocks.identify(this)[0]

    if isinstance(name, (str)):
        click.echo(f"({number}) {name}")


@cli_rocks.command()
@click.argument("this", default="")
@click.option(
    "-m", "--minimal", is_flag=True, help="Reduce output to basic information."
)
def info(this, minimal):
    """Print ssoCard of minor body.

    Parameters
    ==========
    this : str, optional
        Minor body name, designation, or number.
        If empty, a selection is prompted.
    """
    if not this:
        _, _, this = rocks.utils.select_sso_from_index()
    else:  # passed identified string, ensure that we know it
        _, _, this = rocks.identify(this)

    if not isinstance(this, str):
        sys.exit()

    ssoCard = rocks.utils.get_ssoCard(this)

    if ssoCard is not None:
        rocks.utils.pretty_print_card(ssoCard, minimal)


@cli_rocks.command()
def properties():
    """Prints the ssoCard JSON keys."""
    keys = sorted(pd.json_normalize(rocks.TEMPLATE).columns, key=len)
    import pprint

    pprint.pprint(keys)


@cli_rocks.command()
def status():
    """Prints the availability of SsODNet:datacloud. """
    import warnings

    warnings.filterwarnings("ignore")

    from rich import print

    Ceres = rocks.Rock(1)

    if hasattr(Ceres, "taxonomy"):
        print(r"[bold green]Datacloud is available.")
    else:
        print(r"[bold red]Datacloud is not available.")



#!/usr/bin/env python
"""Implement the Rock class and other core rocks functionality.
"""
import copy
from functools import singledispatch
from types import SimpleNamespace
import warnings

import numpy as np
import pandas as pd
from tqdm import tqdm

import rocks


class Rock:
    "For space rocks."

    def __init__(self, identifier, ssoCard=None, datacloud=[], skip_id_check=False):
        """Identify a minor body  and retrieve its properties from SsODNet.

        Parameters
        ==========
        identifier : str, int, float
            Identifying asteroid name, designation, or number
        ssoCard : dict
            Optional previously acquired ssoCard.
        datacloud : list of str
            List of additional catalogues to retrieve from datacloud.
            Default is no additional catalogues.

        Returns
        =======
        rocks.core.Rock
            An asteroid class instance, with its properties as attributes.

        Notes
        =====
        If the asteroid could not be identified, the name and number are None
        and no further attributes are set.

        Example
        =======
        >>> from rocks.core import Rock
        >>> ceres = Rock('ceres')
        >>> ceres.taxonomy.class_
        'C'
        >>> ceres.taxonomy.shortbib
        'DeMeo+2009'
        >>> ceres.diameter
        848.4
        >>> ceres.diameter.unit
        'km'
        """
        if isinstance(datacloud, str):
            datacloud = [datacloud]

        # Identify minor body
        if not skip_id_check:
            self.name, self.number, self.id = rocks.resolve.identify(identifier)[0]
        else:
            self.id = identifier

        if not isinstance(self.id, str):
            return

        ssoCard = ssoCard if ssoCard is not None else rocks.utils.get_ssoCard(self.id)

        if ssoCard is None:
            # Every object needs a name and a number
            if not hasattr(self, "name"):
                self.name, self.number = rocks.resolve.identify(self.id)
            # No data to update
            ssoCard = {}

        # Fill attributes from argument, cache, or query
        ssoCard = rocks.utils.sanitize_keys(ssoCard)
        attributes = copy.deepcopy(rocks.TEMPLATE)
        attributes = rocks.utils.update_ssoCard(attributes, ssoCard)

        # Add JSON keys as attributes, mapping to the appropriate type
        for attribute in attributes.keys():
            setattr(
                self,
                attribute,
                cast_types(attributes[attribute]),
            )

        if self.number == 0:
            self.number = np.nan

        # Set uncertainties and values
        self.__add_metadata()

        # Add datacloud list attributes
        for catalogue in datacloud:
            self.__add_datacloud_catalogue(catalogue)

            # Add preferred entries
            getattr(
                self, rocks.utils.DATACLOUD_META[catalogue]["attr_name"]
            ).select_preferred(catalogue)

    def __hash__(self):
        return hash(self.name)

    def __repr__(self):
        return (
            self.__class__.__qualname__
            + f"(number={self.number!r}, name={self.name!r})"
        )

    def __str__(self):
        return f"({self.number}) {self.name}"

    def __eq__(self, other):
        if self.__class__ is other.__class__:
            return self.name == other.name
        return NotImplemented

    def __lt__(self, other):
        if self.__class__ is other.__class__:
            return (self.number, self.name) < (other.number, other.name)
        return NotImplemented  # pragma: no cover

    def __le__(self, other):
        if self.__class__ is other.__class__:
            return (self.number, self.name) <= (other.number, other.name)
        return NotImplemented  # pragma: no cover

    def __gt__(self, other):
        if self.__class__ is other.__class__:
            return (self.number, self.name) > (other.number, other.name)
        return NotImplemented  # pragma: no cover

    def __ge__(self, other):
        if self.__class__ is other.__class__:
            return (self.number, self.name) >= (other.number, other.name)
        return NotImplemented  # pragma: no cover

    def __getattr__(self, name):
        """Implement attribute shortcuts.
        Gets called if getattribute fails."""
        # Implements shortcuts: omission of parameters.physical/dynamical
        if name in rocks.SHORTCUTS["physical"]:
            return getattr(self.parameters.physical, name)
        elif name in rocks.SHORTCUTS["dynamical"]:
            return getattr(self.parameters.dynamical, name)
        raise AttributeError(f"Unknown attribute {name}")

    def __add_metadata(self):
        """docstring for __add_metadata"""
        for meta, target in rocks.META_MAPPING.items():
            try:
                setattr(
                    rocks.utils.rgetattr(self, target),
                    "uncertainty",
                    rocks.utils.rgetattr(self, meta),
                )
            except AttributeError:
                # some unit paths are currently ill-defined
                pass
            try:
                setattr(
                    rocks.utils.rgetattr(self, target),
                    "unit",
                    rocks.utils.rgetattr(self, meta.replace("uncertainty", "unit")),
                )
            except AttributeError:
                # some unit paths are currently ill-defined
                pass

    def __add_datacloud_catalogue(self, catalogue):
        """docstring for __add_datacloud_catalogue"""
        if not hasattr(getattr(self, "datacloud"), catalogue):
            warnings.warn(f"Unknown datacloud catalogue requested: {catalogue}")
            return

        catalogue_dict = rocks.utils.retrieve_catalogue(
            getattr(getattr(self, "datacloud"), catalogue)
        )

        if catalogue_dict[self.id]["datacloud"] is None:
            setattr(
                self,
                rocks.utils.DATACLOUD_META[catalogue]["attr_name"],
                None,
            )
            return

        catalogue_list = catalogue_dict[self.id]["datacloud"][catalogue]
        catalogue_list = [rocks.utils.sanitize_keys(dict_) for dict_ in catalogue_list]

        setattr(
            self,
            rocks.utils.DATACLOUD_META[catalogue]["attr_name"],
            cast_types(catalogue_list),
        )


class stringParameter(str):
    """For minor body parameters which are strings, e.g. taxonomy."""

    def __new__(self, value):
        return str.__new__(self, value)

    def __init__(self, value):
        str.__init__(value)


class floatParameter(float):
    """For minor body parameters which are floats, e.g. albedo.

    Allows to assign attributes.
    """

    def __new__(self, value):
        return float.__new__(self, value)

    def __init__(self, value):
        float.__init__(value)


class intParameter(int):
    """For minor body parameters which are floats, e.g. number.

    Allows to assign attributes.
    """

    def __new__(self, value):
        return int.__new__(self, value)

    def __init__(self, value):
        float.__init__(value)


class propertyCollection(SimpleNamespace):
    """For collections of data, e.g. taxonomy -> class, method, shortbib.

    Collections of float properties have plotting and averaging methods.
    """

    #  def __repr__(self):
    #  return self.__class__.__qualname__ + json.dumps(self.__dict__, indent=2)

    #  def __str__(self):
    #  return self.__class__.__qualname__ + json.dumps(self.__dict__, indent=2)

    def __len__(self):

        _value_sample = list(self.__dict__.values())[0]

        if isinstance(_value_sample, list):
            return len(_value_sample)
        else:
            return 1

    def __iter__(self):
        self._iter_index = -1
        return self

    def __next__(self):
        self._iter_index += 1

        if len(self) == 1 and self._iter_index == 0:
            return self
        elif self._iter_index < len(self):
            return SimpleNamespace(
                **dict(
                    (k, v[self._iter_index])
                    for k, v in self.__dict__.items()
                    if k != "_iter_index"
                )
            )
        raise StopIteration

    def scatter(self, prop_name, **kwargs):
        return rocks.plots.scatter(self, prop_name, **kwargs)

    def hist(self, prop_name, **kwargs):
        return rocks.plots.hist(self, prop_name, **kwargs)

    def select_preferred(self, prop_name):
        """Select the preferred values based on the observation methods.

        Parameters
        ==========
        prop_name : str
            The property to rank.

        Returns
        =======
        list of bool
            Entry "preferred" in propertyCollection, True if preferred, else False
        """
        if prop_name == "diamalbedo":
            setattr(
                self,
                "preferred_albedo",
                rocks.properties.rank_properties("albedo", self),
            )
            setattr(
                self,
                "preferred_diameter",
                rocks.properties.rank_properties("diameter", self),
            )
            setattr(
                self,
                "preferred",
                [
                    any([di, ai])
                    for di, ai in zip(self.preferred_diameter, self.preferred_albedo)
                ],
            )
        else:
            self.preferred = rocks.properties.rank_properties(prop_name, self)


class listSameTypeParameter(list):
    """For several measurements of a single parameters of any type
    in datcloud catalogues.
    """

    def __init__(self, data):
        """Construct list which allows for assigning attributes.

        Parameters
        ==========
        data : iterable
            The minor body data from datacloud.
        """
        self.datatype = self.__get_type(data)

        if self.datatype is not None:
            list.__init__(self, [self.datatype(d) for d in data])
        else:
            list.__init__(self, [None for d in data])

    def __get_type(self, values):
        """Infers type from str variable."""
        if not values:
            return None
        else:
            try:
                values = [float(value) for value in values]
                if all(
                    [
                        value.is_integer()
                        and "e" not in str(value)
                        and "E" not in str(value)
                        for value in values
                    ]
                ):
                    return int
                else:
                    return float
            except ValueError:
                return str

    def weighted_average(self, errors=None, preferred=[]):
        """Compute weighted average of float-type parameters.

        Parameters
        ==========
        errors : list of floats, np.ndarraya of floats
            Optional list of associated uncertainties. Default is unit
            unceratinty.
        preferred : list of bools
            Compute average only from values where preferred is True.

        Returns
        ======
        (float, float)
        Weighted average and its uncertainty.
        """
        if self.datatype is not float:
            raise TypeError("Property is not of type float.")

        observable = np.array(self)

        # Make uniform weights in case no errors are provided
        if errors is None:
            warnings.warn("No error provided, using uniform weights.")
            errors = np.ones(len(self))
        else:
            # Remove measurements where the error is zero
            errors = np.array(errors)

        if preferred:
            observable = observable[preferred]
            errors = errors[preferred]

        return rocks.utils.weighted_average(observable, errors)


def rocks_(identifier, datacloud=[], progress=False):
    """Create multiple Rock instances via POST request.

    Parameters
    ==========
    identifier : list of str, list of int, list of float, np.array, pd.Series
        An iterable containing minor body identifiers.
    datacloud : list of str
        List of additional catalogues to retrieve from datacloud. Default is
        [], no additional data.
    progress : bool
        Show progress of instantiation. Default is False.

    Returns
    =======
    list of rocks.core.Rock
        A list of Rock instances
    """
    if isinstance(identifier, pd.Series):
        identifier = identifier.values

    # Ensure we know these objects
    ids = [id_ for _, _, id_ in rocks.resolve.identify(identifier)]

    ssoCards = rocks.utils.get_ssoCards(ids, progress=progress)

    if progress:
        progressbar = tqdm(desc="Building rocks", total=len(ids))

    rocks_ = []

    for id_, ssoCard in zip(ids, ssoCards):
        rocks_.append(Rock(id_, ssoCard, datacloud, skip_id_check=True))
        if progress:
            progressbar.update()
    return rocks_


@singledispatch
def cast_types(value):
    return value


@cast_types.register(dict)
def _cast_dict(value):
    return propertyCollection(
        **{
            k: cast_types(v) if isinstance(v, dict) else __TYPES[type(v)](v)
            for k, v in value.items()
        }
    )


@cast_types.register(list)
def _cast_list(list_):
    """Turn lists of dicts into a dict of lists."""
    #  breakpoint()
    if len(list_) == 1:
        return cast_types(list_[0])
    return propertyCollection(
        **{k: listSameTypeParameter([dic[k] for dic in list_]) for k in list_[0]}
    )


__TYPES = {
    type(None): lambda v: None,
    None: lambda v: None,
    int: intParameter,
    str: stringParameter,
    float: floatParameter,
    dict: propertyCollection,
    list: _cast_list,
}
